/******************************************************************/
/** Ce fichier definit les fonctions pour la conversions des     **/
/** formats d'images.                                            **/
/******************************************************************/

/**** Les fichiers d'inclusion ****/

#include <stdlib.h>
#include <string.h>
#include <gd.h>

/**** Les fonctions de conversion ****/

static
unsigned char
citb(int i)
{
  if (i >= 256)
    return 255;
  if (i < 0)
    return 0;
  return i;
}

void
traite_yuv420p(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  const unsigned char *u, *v, *bu, *bv;
  unsigned int dx, dy, uvc, buvc;
  
  u = src + x * y;
  v = u + (x / 2) * (y / 2);
  
  buvc = 0;
  for (dy = 0; dy < y; dy++)
  {
    bu = u;
    bv = v;
    uvc = 0;
    
    for (dx = 0; dx < x; dx++)
    {
      *dst++ = citb((int) *src + 1.140 * ((int) *v - 128));
      *dst++ = citb((int) *src - 0.396 * ((int) *u - 128) - 0.581 * ((int) *v - 128));
      *dst++ = citb((int) *src++ + 2.029 * ((int) *u - 128));
      
      uvc++;
      if (uvc >= 2)
      {
        uvc = 0;
        u++;
        v++;
      }
    }

    buvc++;
    if (buvc < 2)
    {
      u = bu;
      v = bv;
    }
    else
      buvc = 0;
  }
}

void
traite_rgb24(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  memcpy(dst, src, size);
}

void
traite_rgb32(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;
  
  dstend = dst + size; 
  while (dst < dstend)
  {
    *dst++ = *src++;
    *dst++ = *src++;
    *dst++ = *src++;
    src++;
  }
}

void
traite_yuyv(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;
  unsigned char u, v;
  int i;
  
  dstend = dst + size;
  
  while (dst < dstend)
  {
    u = src[1];
    v = src[3];

    for (i = 0; i < 2; i++)
    {
      *dst++ = citb((int) *src + 1.140 * ((int) v - 128));
      *dst++ = citb((int) *src - 0.396 * ((int) u - 128) - 0.581 * ((int) v - 128));
      *dst++ = citb((int) *src + 2.029 * ((int) u - 128));
      src += 2;
    }
  }
}

void
traite_yuv420(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;
  unsigned char *u, *v;
  int ypl, bpl;
  int i, j, k, l;

  dstend = dst + size;
  
  u = src + 4;
  bpl = x * 1.5;
  v = u + bpl;
  ypl = x / 4;
  
  /* YYYY UU YYYY UU
   * YYYY VV YYYY VV */
  while (dst < dstend)
  {
    for (l = 0; l < 2; l++)
    {
      for (k = 0; k < ypl; k++)
      {
        for (j = 0; j < 2; j++)
        {
          for (i = 0; i < 2; i++)
          {
            *dst++ = citb((int) *src + 1.140 * ((int) *v - 128));
            *dst++ = citb((int) *src - 0.396 * ((int) *u - 128) - 0.581 * ((int) *v - 128));
            *dst++ = citb((int) *src + 2.029 * ((int) *u - 128));
            src++;
          }
          u++;
          v++;
        }
        src += 2;
        u += 4;
        v += 4;
      }
      u -= bpl;
      v -= bpl;
    }
    u += bpl * 2;
    v += bpl * 2;
  }
}

void
traite_rgb565(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;

  dstend = dst + size;

  while (dst < dstend)
  {
    /* rrrrrggg gggbbbbb -> gggbbbbb rrrrrggg */
    *dst++ = (src[1] & 0xf8) | ((src[1] >> 3) & 0x07);
    *dst++ = (src[1] << 5) | ((src[0] >> 3) & 0x1c) | ((src[0] >> 5) & 0x03);
    *dst++ = (src[0] << 3) | (src[0] & 0x07);
    src += 2;
  }
}

void
traite_rgb555(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;

  dstend = dst + size;

  while (dst < dstend)
  {
    /* 0rrrrrgg gggbbbbb -> gggbbbbb 0rrrrrgg */
    *dst++ = ((src[1] << 1) & 0xf8) | ((src[1] >> 2) & 0x07);
    *dst++ = (src[1] << 6) | ((src[0] >> 2) & 0x38) | ((src[0] >> 5) & 0x07);
    *dst++ = (src[0] << 3) | (src[0] & 0x07);
    src += 2;
  }
}

#define bayer_R(tab,x,y,width) (tab[0 + 3 * ((x) + width * (y))])
#define bayer_G(tab,x,y,width) (tab[1 + 3 * ((x) + width * (y))])
#define bayer_B(tab,x,y,width) (tab[2 + 3 * ((x) + width * (y))])

#define bayer_RGB(tab,x,y,width) (tab[(x) + width * (y)])

static void bayer_copy(unsigned char *src, unsigned char *dst, int x, int y, int width)
{

  bayer_G(dst,x + 0, y + 0, width) = bayer_RGB(src, x + 0, y + 0, width);
  bayer_G(dst,x + 1, y + 1, width) = bayer_RGB(src, x + 1, y + 1, width);
  bayer_G(dst,x + 0, y + 1, width) = bayer_G(dst,x + 1, y + 0, width) = ((unsigned long)bayer_RGB(src, x + 0, y + 0, width) + (unsigned long)bayer_RGB(src, x + 1, y + 1, width)) / 2;
  bayer_R(dst,x + 0, y + 0, width) = bayer_R(dst,x + 1, y + 0, width) = bayer_R(dst,x + 1, y + 1, width) = bayer_R(dst,x + 0, y + 1, width) = bayer_RGB(src, x + 0, y + 1, width);
  bayer_B(dst,x + 1, y + 1, width) = bayer_B(dst,x + 0, y + 0, width) = bayer_B(dst,x + 0, y + 1, width) = bayer_B(dst,x + 1, y + 0, width) = bayer_RGB(src, x + 1, y + 0, width);
}

static void bayer_bilinear(unsigned char *src, unsigned char *dst, int x, int y, int width)
{
  bayer_R(dst,x + 0, y + 0, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 1, width) + (unsigned long)bayer_RGB(src,x + 0, y - 1, width)) / 2;
  bayer_G(dst,x + 0, y + 0, width) = bayer_RGB(src,x + 0, y + 0, width);
  bayer_B(dst,x + 0, y + 0, width) = ((unsigned long)bayer_RGB(src,x - 1, y + 0, width) + (unsigned long)bayer_RGB(src,x + 1, y + 0, width)) / 2;

  bayer_R(dst,x + 0, y + 1, width) = bayer_RGB(src,x + 0, y + 1, width);
  bayer_G(dst,x + 0, y + 1, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 0, width) + (unsigned long)bayer_RGB(src,x + 0, y + 2, width)
       + (unsigned long)bayer_RGB(src,x - 1, y + 1, width) + (unsigned long)bayer_RGB(src,x + 1, y + 1, width)) / 4;
  bayer_B(dst,x + 0, y + 1, width) = ((unsigned long)bayer_RGB(src,x + 1, y + 0, width) + (unsigned long)bayer_RGB(src,x - 1, y + 0, width)
       + (unsigned long)bayer_RGB(src,x + 1, y + 2, width) + (unsigned long)bayer_RGB(src,x - 1, y + 2, width)) / 4;

  bayer_R(dst,x + 1, y + 0, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 1, width) + (unsigned long)bayer_RGB(src,x + 2, y + 1, width)
       + (unsigned long)bayer_RGB(src,x + 0, y - 1, width) + (unsigned long)bayer_RGB(src,x + 2, y - 1, width)) / 4;
  bayer_G(dst,x + 1, y + 0, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 0, width) + (unsigned long)bayer_RGB(src,x + 2, y + 0, width)
       + (unsigned long)bayer_RGB(src,x + 1, y - 1, width) + (unsigned long)bayer_RGB(src,x + 1, y + 1, width)) / 4;
  bayer_B(dst,x + 1, y + 0, width) = bayer_RGB(src,x + 1, y + 0, width);

  bayer_R(dst,x + 1, y + 1, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 1, width) + (unsigned long)bayer_RGB(src,x + 2, y + 1, width)) / 2;
  bayer_G(dst,x + 1, y + 1, width) = bayer_RGB(src,x + 1, y + 1, width);
  bayer_B(dst,x + 1, y + 1, width) = ((unsigned long)bayer_RGB(src,x + 1, y + 0, width) + (unsigned long)bayer_RGB(src,x + 1, y + 2, width)) / 2;
}


void
traite_gbrg(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  int i, j;

  for (i = 0; i < x; i += 2)
    for (j = 0; j < y; j += 2)
      if (i == 0 || j == 0 || i == x - 2 || j == y - 2)
        bayer_copy(src, dst, i, j, x);
      else
        bayer_bilinear(src, dst, i, j, x);
}


static int verify_jpeg_dht(unsigned char *src,int size_src,
                           unsigned char **dst,int *size_dst)
{
/* This function is based on a patch provided by Scott J. Bertin. */

static unsigned char dht[] =
{
  0xff, 0xc4, 0x01, 0xa2, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01,
  0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
  0x09, 0x0a, 0x0b, 0x10, 0x00, 0x02, 0x01, 0x03, 0x03, 0x02,
  0x04, 0x03, 0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7d,
  0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12, 0x21, 0x31,
  0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32,
  0x81, 0x91, 0xa1, 0x08, 0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52,
  0xd1, 0xf0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
  0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a,
  0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45,
  0x46, 0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57,
  0x58, 0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
  0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x83,
  0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93, 0x94,
  0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
  0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
  0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
  0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8,
  0xd9, 0xda, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8,
  0xe9, 0xea, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
  0xf9, 0xfa, 0x01, 0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01,
  0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a,
  0x0b, 0x11, 0x00, 0x02, 0x01, 0x02, 0x04, 0x04, 0x03, 0x04,
  0x07, 0x05, 0x04, 0x04, 0x00, 0x01, 0x02, 0x77, 0x00, 0x01,
  0x02, 0x03, 0x11, 0x04, 0x05, 0x21, 0x31, 0x06, 0x12, 0x41,
  0x51, 0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81, 0x08, 0x14,
  0x42, 0x91, 0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
  0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34, 0xe1, 0x25,
  0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26, 0x27, 0x28, 0x29, 0x2a,
  0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45, 0x46,
  0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
  0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a,
  0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x82, 0x83,
  0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93, 0x94,
  0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
  0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
  0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
  0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8,
  0xd9, 0xda, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
  0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa
};
unsigned char *p, *i=NULL;

/* By default we simply return the source image. */
*dst=src;
*size_dst=size_src;

/* Scan for an existing DHT segment or the first SOS segment. */
for(p=src+2;p-src<size_src-3 && i==NULL; )
{
  if(*(p++)!=0xFF) continue;
  
  if(*p==0xD9) break;           /* JPEG_EOI */
  if(*p==0xC4) return(0);       /* JPEG_DHT */
  if(*p==0xDA && !i) i=p-1; /* JPEG_SOS */
  
  /* Move to next segment. */
  p += (p[1] << 8)+p[2];
}

/* If no SOS was found, insert the DHT directly after the SOI. */
if(i==NULL) i=src+2;

*size_dst=size_src+sizeof(dht);
*dst=malloc(*size_dst);
if(!*dst) { perror("verify_jpeg_dht.malloc"); exit(-1); }

/* Copy the JPEG data, inserting the DHT segment. */
memcpy((p=*dst),src,i-src);
memcpy((p+=i-src),dht,sizeof(dht));
memcpy((p+=sizeof(dht)),i,size_src-(i-src));

return(1);
}

void
traite_jpeg(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
int size_new;
unsigned char *img_new=NULL;
gdImage *img;
int i;
  
/* MJPEG data may lack the DHT segment required for decoding... */
i=verify_jpeg_dht(src,size,&img_new,&size_new);
  
img=gdImageCreateFromJpegPtr(size_new,img_new);
if(i==1) free(img_new);
if(!img) return;
  
int x0,y0;
for(y0=0;y0<y;y0++)
  for(x0=0;x0<x;x0++)
    {
      int c=gdImageGetPixel(img,x0,y0);
      *dst++ = (c & 0xFF0000) >> 16;
      *dst++ = (c & 0x00FF00) >> 8;
      *dst++ = (c & 0x0000FF);
    }
gdImageDestroy(img);
}

